use std::{
    collections::VecDeque,
    fmt::{Debug, Display},
    fs::File,
    io::Write,
    mem::size_of,
    ops::{Bound, Deref, DerefMut, RangeBounds},
    os::unix::fs::FileExt,
    path::{Path, PathBuf},
};

use crc::{Crc, CRC_32_MPEG_2};
use zerocopy::{
    little_endian::{U32, U64},
    AsBytes, FromBytes, FromZeroes, Unaligned,
};

const CRC: Crc<u32> = Crc::<u32>::new(&CRC_32_MPEG_2);

#[cfg(test)]
mod tests;

#[derive(Debug)]
pub struct LogConfig {
    pub segment_index_size: u32,
    pub segment_data_size: u32,
    pub check_crc_on_read: bool,
}

#[derive(Debug)]
struct ConfigInvalidError {
    segment_index_size: Option<u32>,
}

impl std::error::Error for ConfigInvalidError {}
impl Display for ConfigInvalidError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if let Some(segment_index_size) = self.segment_index_size {
            writeln!(
                f,
                "segment index size too small. is {} but needs to be at least {}",
                segment_index_size,
                LogConfig::MINIMUM_SEGMENT_INDEX_SIZE
            )?;
        }
        Ok(())
    }
}

impl LogConfig {
    const MINIMUM_SEGMENT_INDEX_SIZE: u32 =
        (size_of::<IndexHeader>() + size_of::<IndexEntry>()) as u32;
    fn valid(&self) -> Result<(), ConfigInvalidError> {
        if self.segment_data_size < LogConfig::MINIMUM_SEGMENT_INDEX_SIZE {
            Err(ConfigInvalidError {
                segment_index_size: Some(self.segment_index_size),
            })
        } else {
            Ok(())
        }
    }
}

#[derive(Debug, FromBytes, AsBytes, FromZeroes, Unaligned)]
#[repr(C)]
struct IndexHeader {
    index_offset: U64,
    start: U32,
    end: U32,
}

impl IndexHeader {
    fn eof(&self) -> u32 {
        size_of::<IndexHeader>() as u32 + self.end.get() * size_of::<IndexEntry>() as u32
    }
}

#[derive(Debug, FromBytes, AsBytes, FromZeroes, Unaligned)]
#[repr(C)]
struct IndexEntry {
    offset: U32,
    size: U32,
    crc: U32,
}

struct Index(File);
struct Data(File);

impl Debug for Index {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("Index").finish()
    }
}
impl Debug for Data {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("Data").finish()
    }
}

impl Deref for Index {
    type Target = File;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
impl DerefMut for Index {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}
impl Deref for Data {
    type Target = File;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
impl DerefMut for Data {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl Index {
    fn get(&self, i: u32) -> std::io::Result<IndexEntry> {
        let offset = size_of::<IndexHeader>() as u64 + i as u64 * size_of::<IndexEntry>() as u64;

        let mut entry = IndexEntry::new_zeroed();
        self.read_exact_at(entry.as_bytes_mut(), offset)?;
        Ok(entry)
    }
    fn write_entry(&mut self, i: u32, entry: IndexEntry) -> std::io::Result<()> {
        let offset = size_of::<IndexHeader>() as u64 + i as u64 * size_of::<IndexEntry>() as u64;

        self.write_all_at(entry.as_bytes(), offset)
    }
}

impl Data {
    fn get_entry(&self, entry: IndexEntry, check_crc: bool) -> std::io::Result<Vec<u8>> {
        let mut buf = vec![0; entry.size.get() as usize];
        let offset = entry.offset.get() as u64;
        self.read_exact_at(&mut buf, offset)?;

        if check_crc {
            let mut digest = CRC.digest();
            digest.update(&buf);
            let crc = digest.finalize();
            if crc != entry.crc.get() {
                return Err(std::io::Error::new(
                    std::io::ErrorKind::InvalidData,
                    "log entry corrupted",
                ));
            }
        }

        Ok(buf)
    }
}

#[derive(Debug)]
struct Segment {
    id: u32,
    header: IndexHeader,
    data_eof: u32,
    index: Index,
    data: Data,
    unflushed_index: bool,
    unflushed_data: bool,
}

#[derive(Debug)]
pub struct MessageLog {
    config: LogConfig,
    directory: PathBuf,
    segments: VecDeque<Segment>,
}

fn open(directory: &Path, id: u32, postfix: &str) -> std::io::Result<File> {
    File::options()
        .read(true)
        .write(true)
        .open(directory.join(format!("{id:010}.{postfix}")))
}

fn create(directory: &Path, id: u32, postfix: &str) -> std::io::Result<File> {
    File::options()
        .read(true)
        .write(true)
        .create(true)
        .open(directory.join(format!("{id:010}.{postfix}")))
}

fn delete(directory: &Path, id: u32, postfix: &str) -> std::io::Result<()> {
    std::fs::remove_file(directory.join(format!("{id:010}.{postfix}")))
}

impl MessageLog {
    pub fn new(directory: PathBuf, config: LogConfig) -> std::io::Result<Self> {
        config
            .valid()
            .map_err(|error| std::io::Error::new(std::io::ErrorKind::Other, error))?;

        let mut segments = VecDeque::new();
        if directory.exists() {
            for entry in directory.read_dir()? {
                let entry = entry?;

                if entry.file_type()?.is_file() {
                    if let Ok(name) = entry.file_name().into_string() {
                        if let (Some(id), Some(".index")) = (
                            name.get(..10).and_then(|id| id.parse::<u32>().ok()),
                            name.get(10..),
                        ) {
                            let index = Index(open(&directory, id, "index")?);
                            let data = Data(open(&directory, id, "data")?);

                            let mut header = IndexHeader::new_zeroed();
                            index.read_exact_at(header.as_bytes_mut(), 0)?;

                            let length = header.end.get();

                            let data_eof = if length == 0 {
                                0
                            } else {
                                let entry = index.get(length - 1)?;
                                entry.offset.get() + entry.size.get()
                            };

                            segments.push_back(Segment {
                                id,
                                header,
                                index,
                                data,
                                data_eof,
                                unflushed_index: false,
                                unflushed_data: false,
                            });
                        }
                    }
                }
            }

            segments
                .make_contiguous()
                .sort_unstable_by_key(|header| header.id);
        } else {
            std::fs::create_dir_all(&directory)?;
        };

        Ok(Self {
            config,
            directory,
            segments,
        })
    }
}

impl Segment {
    fn is_empty(&self) -> bool {
        self.header.end.get() == self.header.start.get()
    }
    fn last_index(&self) -> Option<u64> {
        if self.is_empty() {
            return None;
        }
        Some(self.header.index_offset.get() + self.header.end.get() as u64 - 1)
    }
    fn first_index(&self) -> Option<u64> {
        if self.is_empty() {
            return None;
        }
        Some(self.header.index_offset.get() + self.header.start.get() as u64)
    }
    fn delete(self, directory: &Path) -> std::io::Result<()> {
        println!("deleting segment {}", self.id);
        drop(self.data);
        drop(self.index);

        delete(directory, self.id, "index")?;
        delete(directory, self.id, "data")?;

        Ok(())
    }
    fn write_header(&mut self) -> std::io::Result<()> {
        self.unflushed_index = true;
        self.index.write_all_at(self.header.as_bytes(), 0)
    }
}

impl MessageLog {
    fn find_segment(&self, index: u64) -> Option<usize> {
        let res = self
            .segments
            .binary_search_by_key(&index, |segment| segment.header.index_offset.get());

        match res {
            Ok(i) => Some(i),
            Err(0) => None,
            Err(i) => Some(i - 1),
        }
    }

    pub fn last_index(&self) -> Option<u64> {
        let last_segment = self.segments.back()?;
        last_segment.last_index().or_else(|| {
            (self.segments.len() > 1).then_some(last_segment.header.index_offset.get() - 1)
        })
    }

    fn add_segment(&mut self) -> std::io::Result<&mut Segment> {
        let index_offset = self.last_index().map_or(0, |index| index + 1);
        let id = self.segments.back().map_or(0, |segment| segment.id + 1);

        let header = IndexHeader {
            index_offset: index_offset.into(),
            start: 0.into(),
            end: 0.into(),
        };

        let index = Index(create(&self.directory, id, "index")?);
        let data = Data(create(&self.directory, id, "data")?);

        index.write_all_at(header.as_bytes(), 0)?;

        self.segments.push_back(Segment {
            id,
            header,
            index,
            data,
            data_eof: 0,
            unflushed_index: true,
            unflushed_data: true,
        });

        Ok(self.last_segment_mut().unwrap())
    }

    fn last_segment_mut(&mut self) -> Option<&mut Segment> {
        self.segments.back_mut()
    }

    pub fn append(&mut self, bytes: impl AsRef<[u8]>) -> std::io::Result<u64> {
        self._append(bytes.as_ref())
    }
    fn _append(&mut self, bytes: &[u8]) -> std::io::Result<u64> {
        if bytes.len() >= u32::MAX as usize {
            // TODO:
            // return Err(std::io::Error::new(std::io::ErrorKind::FileTooLarge, "message to large for log"));
            return Err(std::io::Error::other("message to large for log"));
        }

        let segment = if self.segments.is_empty() {
            self.add_segment()?
        } else {
            let segment = self.segments.back_mut().unwrap();

            let size_remaining_in_index =
                self.config
                    .segment_index_size
                    .saturating_sub(segment.header.eof()) as usize;

            let size_remaining_in_data = self
                .config
                .segment_data_size
                .saturating_sub(segment.data_eof) as usize;

            if bytes.len() <= size_remaining_in_data
                && size_of::<IndexEntry>() <= size_remaining_in_index
            {
                segment
            } else {
                self.add_segment()?
            }
        };

        let offset = segment.data_eof.into();
        segment.data.write_all_at(bytes, segment.data_eof as u64)?;
        segment.data_eof += bytes.len() as u32;
        segment.unflushed_data = true;

        let relative_index = segment.header.end.get();

        let mut digest = CRC.digest();
        digest.update(bytes);

        segment.index.write_entry(
            relative_index,
            IndexEntry {
                offset,
                size: (bytes.len() as u32).into(),
                crc: digest.finalize().into(),
            },
        )?;

        let index_offset = segment.header.index_offset.get();
        segment.header.end += U32::from(1);
        segment.write_header()?;

        Ok(index_offset + relative_index as u64)
    }

    pub fn get(&self, index: u64) -> std::io::Result<Option<Vec<u8>>> {
        let Some(segment_index) = self.find_segment(index) else {
            return Ok(None);
        };
        let segment = &self.segments[segment_index];
        let relative_index = index - segment.header.index_offset.get();
        let entry = segment.index.get(relative_index as u32)?;

        segment
            .data
            .get_entry(entry, self.config.check_crc_on_read)
            .map(Some)
    }

    pub fn get_range(&self, range: impl RangeBounds<u64>) -> LogIterator<'_> {
        if self.segments.is_empty() {
            return LogIterator::Empty;
        }

        let start = range.start_bound().cloned();
        let end = range.end_bound().cloned();

        // if the first segment would be before the first it's the same as starting at the first segment
        let first_segment = match start {
            Bound::Included(index) => self.find_segment(index),
            Bound::Excluded(index) => self.find_segment(index + 1),
            Bound::Unbounded => None,
        }
        .unwrap_or(0);

        let end_segment = match end {
            Bound::Included(index) => self.find_segment(index),
            Bound::Excluded(index) => self.find_segment(index + 1),
            Bound::Unbounded => Some(self.segments.len() - 1),
        };

        // if the last segment would be before the first segment the iterator is empty
        let Some(last_segment) = end_segment else {
            return LogIterator::Empty;
        };

        let segment = &self.segments[first_segment];

        let index_offset = segment.header.index_offset.get();
        let start_offset = index_offset + segment.header.start.get() as u64;
        let index = match start {
            Bound::Included(index) => (index).max(start_offset),
            Bound::Excluded(index) => (index + 1).max(start_offset),
            Bound::Unbounded => start_offset,
        };

        let segment_over = segment
            .last_index()
            .map_or(true, |last_index| index > last_index);

        if segment_over {
            return LogIterator::Empty;
        }

        let index_in_segment = (index - index_offset) as u32;

        LogIterator::Data {
            log: self,
            end,
            last_segment,
            segment_index: first_segment,
            index_in_segment,
        }
    }
    pub fn length(&self) -> u64 {
        if self.is_empty() {
            0
        } else {
            let front = self.segments.front().unwrap();
            let back = self.segments.back().unwrap();

            let start = front
                .first_index()
                .unwrap_or(front.header.index_offset.get());
            let end = back
                .last_index()
                .unwrap_or(back.header.index_offset.get().saturating_sub(1));

            end + 1 - start
        }
    }

    pub fn is_empty(&self) -> bool {
        self.segments.is_empty() || (self.segments.len() == 1 && self.segments[0].is_empty())
    }

    pub fn truncate(&mut self, from: u64) -> std::io::Result<()> {
        if self.is_empty() {
            return Ok(());
        }

        let from_i = self.find_segment(from).unwrap_or(0);

        let segments = self
            .segments
            .drain((Bound::Excluded(from_i), Bound::Unbounded));

        for segment in segments.rev() {
            segment.delete(&self.directory)?;
        }

        let segment = &mut self.segments[from_i];

        let local_index =
            if let Some(local_index) = from.checked_sub(segment.header.index_offset.get()) {
                local_index as u32
            } else {
                segment.header.index_offset = from.into();
                0
            };

        segment.header.end = local_index.into();
        segment.header.start = segment.header.start.get().min(local_index).into();
        segment.write_header()?;

        segment.data_eof = if segment.is_empty() {
            0
        } else {
            let entry = segment.index.get(segment.header.end.get() - 1)?;
            entry.offset.get() + entry.size.get()
        };

        Ok(())
    }

    pub fn trim(&mut self, to: u64) -> std::io::Result<()> {
        if self.is_empty() {
            return Ok(());
        }

        let Some(to_i) = self.find_segment(to) else {
            return Ok(());
        };

        let segments = self
            .segments
            .drain((Bound::Unbounded, Bound::Excluded(to_i)));

        let len = segments.len();
        for segment in segments {
            segment.delete(&self.directory)?;
        }

        let segment = &mut self.segments[to_i - len];

        let local_index = to.saturating_sub(segment.header.index_offset.get()) as u32;

        segment.header.start = local_index.into();
        segment.write_header()?;

        Ok(())
    }

    pub fn flush(&mut self) -> std::io::Result<()> {
        for segment in &mut self.segments {
            if segment.unflushed_data {
                segment.data.0.flush()?;
            }

            if segment.unflushed_index {
                segment.index.0.flush()?;
            }
        }

        Ok(())
    }
}

pub enum LogIterator<'log> {
    Empty,
    Data {
        log: &'log MessageLog,
        end: Bound<u64>,
        last_segment: usize,
        segment_index: usize,
        index_in_segment: u32,
    },
}

impl<'log> Iterator for LogIterator<'log> {
    type Item = std::io::Result<Vec<u8>>;

    fn next(&mut self) -> Option<Self::Item> {
        match self {
            LogIterator::Empty => None,
            LogIterator::Data {
                log,
                end,
                last_segment,
                segment_index,
                index_in_segment,
            } => {
                if segment_index > last_segment {
                    *self = LogIterator::Empty;
                    return None;
                }

                let mut segment = &log.segments[*segment_index];

                let segment_over = *index_in_segment >= segment.header.end.get();

                if segment_over {
                    *segment_index += 1;

                    if segment_index > last_segment {
                        *self = LogIterator::Empty;
                        return None;
                    }

                    segment = &log.segments[*segment_index];
                    *index_in_segment = 0;
                }

                let local_index = *index_in_segment;
                *index_in_segment += 1;
                let index = segment.header.index_offset.get() + local_index as u64;
                let in_bounds = match end {
                    Bound::Included(end) => index <= *end,
                    Bound::Excluded(end) => index < *end,
                    Bound::Unbounded => true,
                };

                if !in_bounds {
                    *self = LogIterator::Empty;
                    return None;
                }

                Some(match segment.index.get(local_index) {
                    Err(err) => return Some(Err(err)),
                    Ok(entry) => segment.data.get_entry(entry, log.config.check_crc_on_read),
                })
            }
        }
    }
}
