use std::{
    ffi::OsString,
    fmt::Debug,
    ops::{Bound, RangeBounds},
    path::Path,
    vec,
};

use crate::{LogConfig, MessageLog};

#[test]
fn basic() {
    let tempdir = tempfile::tempdir().unwrap();

    let mut log = MessageLog::new(
        tempdir.path().to_owned(),
        LogConfig {
            segment_index_size: 64,
            segment_data_size: 32,
            check_crc_on_read: true,
        },
    )
    .unwrap();

    let data = vec![
        vec![0],
        vec![1],
        vec![2],
        vec![3],
        vec![4],
        vec![5],
        vec![6],
        vec![7],
        vec![8],
    ];

    assert_eq!(log.last_index(), None);
    for (i, element) in data.iter().enumerate() {
        assert_eq!(log.append(element).unwrap(), i as u64);
        assert_eq!(log.last_index(), Some(i as u64));
    }

    for (i, element) in data.iter().enumerate() {
        assert_eq!(log.get(i as u64).unwrap().as_ref(), Some(element));
    }

    #[track_caller]
    fn assert_range<R>(log: &MessageLog, data: &[Vec<u8>], range: R)
    where
        R: RangeBounds<u64> + Debug + Clone,
    {
        fn convert_bound(bound: Bound<&u64>) -> Bound<usize> {
            match bound {
                Bound::Included(x) => Bound::Included(*x as usize),
                Bound::Excluded(x) => Bound::Excluded(*x as usize),
                Bound::Unbounded => Bound::Unbounded,
            }
        }

        let got = log
            .get_range(range.clone())
            .collect::<std::io::Result<Vec<_>>>()
            .unwrap();

        let expected = &data[(
            convert_bound(range.start_bound()),
            convert_bound(range.end_bound()),
        )];

        assert_eq!(
            got,
            expected,
            "failed for range {range:?} got {} elements, expected {}",
            got.len(),
            expected.len()
        );
    }

    assert_range(&log, &data, ..);

    let n = data.len() as u64 + 1;
    for i in 0..n {
        assert_range(&log, &data, ..i);
        assert_range(&log, &data, i..);

        for j in i..n {
            assert_range(&log, &data, i..j);
        }
    }

    // can't store messages larger than 4GiB
    assert_eq!(log.append([1; 128]).unwrap(), 9);
    assert_eq!(log.append([2; 30]).unwrap(), 10);
    assert_eq!(log.append([3; 30]).unwrap(), 11);
    assert!(log.append(&vec![0; u32::MAX as usize + 1]).is_err());

    log.truncate(8).unwrap();
    assert_eq!(log.last_index(), Some(7));
    assert!(!log.is_empty());
    assert_eq!(log.length(), 8);

    log.truncate(1).unwrap();
    assert_eq!(log.last_index(), Some(0));
    assert!(!log.is_empty());
    assert_eq!(log.length(), 1);
    assert_eq!(log.append([0; 30]).unwrap(), 1);

    log.truncate(0).unwrap();
    assert!(log.is_empty());

    assert_eq!(log.append([0; 30]).unwrap(), 0);
    assert_eq!(log.length(), 1);

    for i in 1..10 {
        assert_eq!(log.append([i as u8; 30]).unwrap(), i);
    }

    log.trim(4).unwrap();

    assert_eq!(log.length(), 6);

    drop(log);

    let mut log = MessageLog::new(
        tempdir.path().to_owned(),
        LogConfig {
            segment_index_size: 100,
            segment_data_size: 300,
            check_crc_on_read: true,
        },
    )
    .unwrap();

    for i in 10..20 {
        assert_eq!(log.append([i as u8; 30]).unwrap(), i);
        assert_eq!(log.length(), i + 1 - 4);
    }

    for (i, entry) in log.get_range(..).enumerate() {
        assert_eq!(entry.unwrap(), &[4 + i as u8; 30]);
    }

    log.trim(20).unwrap();

    assert_eq!(log.append([0; 30]).unwrap(), 20);

    log.flush().unwrap();

    list(tempdir.path());
}

fn list(dir: &Path) {
    let files = dir
        .read_dir()
        .unwrap()
        .map(|entry| Ok(entry?.file_name().to_os_string()))
        .collect::<std::io::Result<Vec<OsString>>>()
        .unwrap();

    std::process::Command::new("wc")
        .arg("-c")
        .args(files)
        .current_dir(dir)
        .spawn()
        .unwrap()
        .wait()
        .unwrap();
}
